! ppm_graph_partition_serial.f90 --- wrapper for single task graph partitioner
!
! Copyright  (C)  2011  Thomas Jahns <jahns@dkrz.de>
!
! Version: 1.0
! Keywords: graph partitioning
! Author: Thomas Jahns <jahns@dkrz.de>
! Maintainer: Thomas Jahns <jahns@dkrz.de>
! URL: https://www.dkrz.de/redmine/projects/scales-ppm
!
! Redistribution and use in source and binary forms, with or without
! modification, are  permitted provided that the following conditions are
! met:
!
! Redistributions of source code must retain the above copyright notice,
! this list of conditions and the following disclaimer.
!
! Redistributions in binary form must reproduce the above copyright
! notice, this list of conditions and the following disclaimer in the
! documentation and/or other materials provided with the distribution.
!
! Neither the name of the DKRZ GmbH nor the names of its contributors
! may be used to endorse or promote products derived from this software
! without specific prior written permission.
!
! THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
! IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
! TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
! PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
! OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
! EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
! PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
! PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
! LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
! NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
! SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
!
!> perform partitioning of graph from serial code
MODULE ppm_graph_partition_serial
  USE iso_c_binding
  USE ppm_base, ONLY: assertion
  USE ppm_extents, ONLY: extent
  USE ppm_graph_csr, ONLY: graph_csr, num_nodes
  USE ppm_set_partition_base, ONLY: partition_assignment
  USE ppm_std_type_kinds, ONLY: i4
  IMPLICIT NONE
  PRIVATE
#include <ppm.inc>
#ifdef HAVE_METIS_V4
  INTEGER, PARAMETER :: ppm_metis_int = c_int
#else
  INTEGER, PARAMETER :: ppm_metis_int = ppm_metis_idx
#endif
#ifdef HAVE_METIS_V4
  INTERFACE
    SUBROUTINE metis_mcpartgraphkway(nvtxs, ncon, xadj, adjncy, vwgt, adjwgt, &
         wgtflag, numflag, nparts, rubvec, options, edgecut, part) BIND(C)
      USE iso_c_binding, ONLY: c_int, c_ptr
      IMPORT :: ppm_metis_idx
      IMPORT :: ppm_metis_real
      INTEGER(ppm_metis_idx), INTENT(in) :: xadj(*), adjncy(*)
      TYPE(c_ptr), VALUE, INTENT(in) :: vwgt, adjwgt
      INTEGER(c_int), INTENT(in) :: nvtxs, wgtflag, numflag, ncon, nparts, options(*)
      REAL(ppm_metis_real), INTENT(in) :: rubvec(ncon)
      INTEGER(c_int), INTENT(out) :: edgecut
      INTEGER(ppm_metis_idx), INTENT(out) :: part(*)
    END SUBROUTINE metis_mcpartgraphkway
  END INTERFACE
#else
  INTERFACE
    SUBROUTINE metis_setdefaultoptions(options) BIND(C)
      IMPORT :: ppm_metis_idx
      INTEGER(ppm_metis_idx), INTENT(out) :: options(*)
    END SUBROUTINE metis_setdefaultoptions
  END INTERFACE
#endif
  INTERFACE
#ifdef HAVE_METIS_V4
    SUBROUTINE metis_partgraphkway(nvtxs, xadj, adjncy, vwgt, adjwgt, &
         wgtflag, numflag, nparts, options, edgecut, part) BIND(C)
      USE iso_c_binding, ONLY: c_int
#else
    SUBROUTINE metis_partgraphkway(nvtxs, ncon, xadj, adjncy, vwgt, vsize, &
         adjwgt, nparts, tpwgts, ubvec, options, edgecut, part) BIND(C)
#endif
      USE iso_c_binding, ONLY: c_ptr
      IMPORT :: ppm_metis_idx, ppm_metis_int
      INTEGER(ppm_metis_idx), INTENT(in) :: xadj(*), adjncy(*)
      TYPE(c_ptr), VALUE, INTENT(in) :: vwgt, adjwgt
      INTEGER(ppm_metis_int), INTENT(in) :: nvtxs, nparts, options(*)
#ifdef HAVE_METIS_V4
      INTEGER(c_int), INTENT(in) :: wgtflag, numflag
#else
      INTEGER(ppm_metis_idx), INTENT(in) :: ncon
      TYPE(c_ptr), VALUE, INTENT(in) :: vsize
      TYPE(c_ptr), VALUE, INTENT(in) :: tpwgts, ubvec
#endif
      INTEGER(ppm_metis_int), INTENT(out) :: edgecut
      INTEGER(ppm_metis_idx), INTENT(out) :: part(*)
    END SUBROUTINE metis_partgraphkway
  END INTERFACE
  PUBLIC :: graph_partition_metis
  INTERFACE graph_partition_metis
    MODULE PROCEDURE graph_partition_metis_base
    MODULE PROCEDURE graph_partition_metis_csr
    MODULE PROCEDURE graph_partition_metis_csr_mweights
  END INTERFACE graph_partition_metis

  CHARACTER(len=*), PARAMETER :: filename = 'ppm_graph_partition_serial.f90'
CONTAINS
  SUBROUTINE graph_partition_metis_base(num_vertices, edge_list_lens, &
       edge_lists, partition_out, num_partitions, &
       imbalance_tolerance, vertex_weights, edge_weights)
    INTEGER(ppm_metis_int), INTENT(in) :: num_vertices
    INTEGER(ppm_metis_idx), INTENT(in) :: edge_list_lens(*)
    INTEGER(ppm_metis_idx), INTENT(in) :: edge_lists(*)
    INTEGER(ppm_metis_idx), INTENT(out) :: partition_out(*)
    INTEGER(ppm_metis_int), INTENT(in) :: num_partitions
    REAL(ppm_metis_real), OPTIONAL, TARGET, INTENT(in) :: imbalance_tolerance(:)
    INTEGER(ppm_metis_idx), OPTIONAL, TARGET, INTENT(in) :: vertex_weights(*)
    INTEGER(ppm_metis_idx), OPTIONAL, TARGET, INTENT(in) :: edge_weights(*)
    TYPE(c_ptr) :: vwgt, adjwgt
#ifdef HAVE_METIS_V4
    INTEGER(c_int) :: wgtflag
    INTEGER(c_int) :: metis_options(0:4), edge_cut

    metis_options = 0
#else
    TYPE(c_ptr) :: vsize = c_null_ptr
    TYPE(c_ptr) :: tpwgts = c_null_ptr, ubvec
    INTEGER(ppm_metis_idx) :: metis_options(0:39), edge_cut

    CALL metis_setdefaultoptions(metis_options)
    ! METIS_OPTION_NUMBERING : use Fortran-style
    metis_options(17) = 1
#endif

    IF (PRESENT(vertex_weights)) THEN
      vwgt = c_loc(vertex_weights(1))
    ELSE
      vwgt = c_null_ptr
    END IF
    IF (PRESENT(edge_weights)) THEN
      adjwgt = c_loc(edge_weights(1))
    ELSE
      adjwgt = c_null_ptr
    END IF
    IF (PRESENT(imbalance_tolerance)) THEN
      CALL assertion(PRESENT(vertex_weights), line=__LINE__, source=filename, &
           msg="when imbalance_tolerance is provided, vertex weights&
           & are also required")
#ifdef HAVE_METIS_V4
      wgtflag = MERGE(1, 0, PRESENT(edge_weights))
      CALL metis_mcpartgraphkway(num_vertices, &
           INT(SIZE(imbalance_tolerance), c_int), edge_list_lens, edge_lists, &
           vwgt, adjwgt, wgtflag, 1, num_partitions, imbalance_tolerance, &
           metis_options, edge_cut, partition_out)
#else
      ubvec = c_loc(imbalance_tolerance(1))
#endif
    ELSE
#ifdef HAVE_METIS_V4
      wgtflag = MERGE(2, 0, PRESENT(vertex_weights))
      wgtflag = IOR(wgtflag, MERGE(1, 0, PRESENT(edge_weights)))
      CALL metis_partgraphkway(num_vertices, edge_list_lens, edge_lists, vwgt, &
           adjwgt, wgtflag, 1, num_partitions, metis_options, edge_cut, &
           partition_out)
#else
      ubvec = c_null_ptr
#endif
    END IF
#ifndef HAVE_METIS_V4
    CALL metis_partgraphkway(num_vertices, 1, edge_list_lens, &
         edge_lists, vwgt, vsize, adjwgt, num_partitions, tpwgts, ubvec, &
         metis_options, edge_cut, partition_out)
#endif
  END SUBROUTINE graph_partition_metis_base

  SUBROUTINE graph_partition_metis_csr(partition, graph, num_partitions, &
       vertex_weights, edge_weights)
    TYPE(partition_assignment), INTENT(out) :: partition
    TYPE(graph_csr), INTENT(in) :: graph
    INTEGER, INTENT(in) :: num_partitions
    INTEGER(ppm_metis_idx), OPTIONAL, INTENT(in) :: vertex_weights(*)
    INTEGER(ppm_metis_idx), OPTIONAL, INTENT(in) :: edge_weights(*)

    INTEGER :: n, wgtflag
    INTEGER(ppm_metis_idx), ALLOCATABLE :: xadj(:), adjncy(:)

    n = num_nodes(graph)
    wgtflag = MERGE(2, 0, PRESENT(vertex_weights))
    wgtflag = IOR(wgtflag, MERGE(1, 0, PRESENT(edge_weights)))
    partition%part_range = extent(1, num_partitions)
    ALLOCATE(partition%assigned(1:n))
    IF (ppm_metis_idx == i4) THEN
      SELECT CASE(wgtflag)
      CASE(0)
        CALL graph_partition_metis_base(n, graph%edges_of_vtx, &
             graph%edges, partition%assigned, num_partitions)
      CASE(1)
        CALL graph_partition_metis_base(n, graph%edges_of_vtx, &
             graph%edges, partition%assigned, &
             num_partitions, edge_weights=edge_weights)
      CASE(2)
        CALL graph_partition_metis_base(n, graph%edges_of_vtx, &
             graph%edges, partition%assigned, &
             num_partitions, vertex_weights=vertex_weights)
      CASE(3)
        CALL graph_partition_metis_base(n, graph%edges_of_vtx, &
             graph%edges, partition%assigned, num_partitions, &
             vertex_weights=vertex_weights, edge_weights=edge_weights)
      END SELECT
    ELSE
      ALLOCATE(xadj(SIZE(graph%edges_of_vtx)), adjncy(SIZE(graph%edges)))
      xadj = INT(graph%edges_of_vtx, ppm_metis_idx)
      adjncy = INT(graph%edges, ppm_metis_idx)

      CALL graph_partition_metis_base(n, xadj, adjncy, partition%assigned, &
           num_partitions, vertex_weights=vertex_weights, &
           edge_weights=edge_weights)
    END IF
  END SUBROUTINE graph_partition_metis_csr

  SUBROUTINE graph_partition_metis_csr_mweights(partition, graph, &
       num_partitions, imbalance_tolerance, vertex_weights, edge_weights)
    TYPE(partition_assignment), INTENT(out) :: partition
    TYPE(graph_csr), INTENT(in) :: graph
    INTEGER, INTENT(in) :: num_partitions
    REAL(ppm_metis_real), INTENT(in) :: imbalance_tolerance(:)
    INTEGER(ppm_metis_idx), INTENT(in) :: vertex_weights(:,:)
    INTEGER(ppm_metis_idx), OPTIONAL, INTENT(in) :: edge_weights(*)

    INTEGER :: n
    INTEGER(ppm_metis_idx), ALLOCATABLE :: xadj(:), adjncy(:)

    n = num_nodes(graph)
    partition%part_range = extent(1, num_partitions)
    ALLOCATE(partition%assigned(1:n))
    IF (ppm_metis_idx == i4) THEN
      IF (PRESENT(edge_weights)) THEN
        CALL graph_partition_metis_base(n, graph%edges_of_vtx, &
             graph%edges, partition%assigned, &
             num_partitions, imbalance_tolerance, vertex_weights, edge_weights)
      ELSE
        CALL graph_partition_metis_base(n, graph%edges_of_vtx, &
             graph%edges, partition%assigned, &
             num_partitions, imbalance_tolerance, vertex_weights)
      END IF
    ELSE
      ALLOCATE(xadj(SIZE(graph%edges_of_vtx)), adjncy(SIZE(graph%edges)))
      xadj = INT(graph%edges_of_vtx, ppm_metis_idx)
      adjncy = INT(graph%edges, ppm_metis_idx)
      IF (PRESENT(edge_weights)) THEN
        CALL graph_partition_metis_base(n, xadj, adjncy, partition%assigned, &
             num_partitions, imbalance_tolerance, vertex_weights, edge_weights)
      ELSE
        CALL graph_partition_metis_base(n, xadj, adjncy, partition%assigned, &
             num_partitions, imbalance_tolerance, vertex_weights)
      END IF
    END IF
  END SUBROUTINE graph_partition_metis_csr_mweights

END MODULE ppm_graph_partition_serial
!
! Local Variables:
! license-project-url: "https://www.dkrz.de/redmine/projects/scales-ppm"
! license-default: "bsd"
! End:
